/*
** interface.h - defines the data structure for an interface
**
** Keith Zaback (krz@wcug.wwu.edu) Apr 98
*/

#ifndef INC_INTERFACE_H
#define INC_INTERFACE_H

#include <sys/socket.h>

#include "config.h"
#include "hash.h"
#include "ipaddr.h"
#include "buffer.h"
#include "fury.h"
#include "pktcache.h"
#include "hdrcache.h"

/* defines for INTERFACE.flags */
#define FURY_IF_NAME_MAX	16
#define FURY_IF_VALID		0x0001
#define FURY_IF_MODEM_LINES	0x0002
#define FURY_IF_LCP_IGNORE	0x0004
#define FURY_IF_TOO_MUCH_DATA	0x0008

#define FURY_IF_TYPE_UNKNOWN	0x0000
#define FURY_IF_TYPE_SERIAL	0x0001
#define FURY_IF_TYPE_ETHER	0x0002
#define FURY_IF_TYPE_UDP	0x0004

#define FURY_FRAMETYPE_UNKNOWN	0x0000
#define FURY_FRAMETYPE_II	0x0001
#define FURY_FRAMETYPE_802_3	0x0002
#define FURY_FRAMETYPE_802_2	0x0003

/* for the options */
#define FOPT_NONE		0x0000
#define FOPT_FULLFRAME		0x0001	/* always start UDP packets with a full frame */
#define FOPT_FORWARD		0x0002	/* do we forward packets */
#define FOPT_SC_RESEND		0x0004	/* be smarter about SC resends */
#define FOPT_QUITING		0x8000	/* we want to quit */

/* an "interface" */
typedef struct stINTERFACE {
	int		type;			/* type of interface */
	int		fd;			/* file descriptor for this interface */
	char		name[FURY_IF_NAME_MAX];	/* name of interface */
	union {
		struct {			/* for ethernet */
#ifdef FURY_ETHERNET
			int		e_frametype;
#ifdef FURY_LINUX
			struct sockaddr	e_sa;		
#endif

#ifdef FURY_BSD			
			BYTE		*e_buffer;
			int		e_buffer_len;
#endif
#endif /* FURY_ETHERNET */
		} e;
		struct {			/* for serial */
			struct termios	s_tio;		/* for restoring terminal attributes */
			int		s_baudrate;	/* what the baudrate is */
			HDRCACHE	s_send_hdrcache;
			HDRCACHE	s_recv_hdrcache;
			HASHTABLE	s_send_pktcache;
			HASHTABLE	s_recv_pktcache;
			BYTE		s_last_senddata[FURY_FRAME_MAX_DATA];
			int		s_last_sendsize;
			BYTE		s_last_recvdata[FURY_FRAME_MAX_DATA];
			int		s_last_recvsize;
			int		s_diff_reset;
		} s;
		struct {			/* for UDP */
			HASHTABLE	u_ipaddrs;	/* table of IP address */
			HASHTABLE	u_hw_to_ip;	/* maps ethernet address to IPs */
			int		u_portnumber;	/* port number we listen to */
			IPADDR_ENTRY	*u_lastrecv;	/* where we last received our stuff */
			int		u_buffer_threshold; /* min packet size */
			int		u_buffer_wait_sec;
			int		u_buffer_wait_usec;
			int		u_buffer_q_max;	/* max wait (calls to process) */
			int		u_options;	/* our IP options */
		} u;
	} data;

	struct timeval	exit_time;		/* time of forced exit */

	int		options;		/* interface options! */
	
	int		frames_sent;		/* how many fury frames sent this "tick" */
	FURYFRAME	furyframe; 		/* current frame we are building */
	
	BUFFER		recvbuf;		/* receive buffer */
	BUFFER		sendbuf;		/* transmit buffer */
	BUFFER		*outbuf;		/* ptr to the buffer to hold outgoing fury stuff */
	
	FURY_CACHE_PTRS	cache_ptrs;		/* ptrs to the cache */

	int		generation;	
	int		generation_max;		/* counter */
	
	int		flags;			/* info */
	int		state;			/* what state it is in */
	int		restart_counter;	/* how many times to try */
	
	int		(*open)(void *interface);
	int		(*close)(void *interface);
	int		(*read)(void *interface);
	int		(*write)(void *interface);
	int		(*process)(void *interface);
	int		(*queue_ipx_frame)(void *from, void *interface, void *ipxframe);
	int		(*timeout_handler)(void *interface);
	int		(*flush)(void *interface);
	int		(*event_handler)(void *interface, int event);
	int		(*prepare_device)(void *interface);
	int		(*ipxframe_processor)(void *fdev, void *ipxframe);
		
	IFACE_STATS	stats;			/* info on this interface */
} INTERFACE;

#endif
