/*
** hdrcache.c - holds the IPX header caching routines
**
** Keith Zaback (krz@wcug.wwu.edu) Apr 98
*/

#include "ipxfury.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/* protos */
int		hdrcache_hash(IPXHEADER *header);


/*
**	HDRCACHE_IPXHEADER - returns ipx header
*/
IPXHEADER *hdrcache_ipxheader(HDRCACHE *cache, int num)
{
	return( &(cache->table[num].ipx_header) );

}


/*
**	HDRCACHE_HASH - the hashing function
*/
int hdrcache_hash(IPXHEADER *header)
{
	int		i;
	int		hash;
	unsigned char	*ptr;
	int		flip;
	
	/* hash starts as 0 */
	hash = 0;
	flip = 0;
	
	/* add up the src dest ipx addresses */
	/* dest comes 1st, then source */
	ptr = (unsigned char *)&header->ipx_dest;
	for( i = 0; i < (sizeof(IPXADDR)*2); i++ ) {
		if( flip == 1 ) {
			hash += *ptr;
		} else {
			hash += (*ptr << (hash & 0x3));
		}
		ptr++;
		
		/* flip */
		flip ^= 1;
	}
	
	/* return the hash value */
	return(hash);
}



/*
**	HDRCACHE_LOOKUP - check for an item in the cache and return its index
*/
int hdrcache_lookup(HDRCACHE *cache, IPXHEADER *header)
{
	int		bno, i;
	HDRCACHE_ENTRY	*entry;
	
	/* calc the bucket number it should be in */
	bno = hdrcache_hash(header) % cache->size;
	
	/* look for the entry, searching to the bottom of the table */
	i = bno;
	entry = &cache->table[i];
	while( i < cache->size ) {	/* go to the bottom of the table */
		if( (entry->flags & CACHE_VALID) == CACHE_VALID ) {
			if( memcmp(&entry->ipx_header, header, sizeof(IPXHEADER)) == 0 ) {
				/* found it! */
				entry->flags        |= CACHE_ACCESSED;	/* we used this one */
				entry->access_remove = CACHE_REMOVE_ACCESS_TICKS;	/* reset the counter */
				entry->used_count++;	/* one more hit */
				return(i);		/* return its index */
			} 
		} else {
			/* empty slot, meaning the entry is not in the table */
			return(-1);
		}
		i++;		/* move to the next entry */
		entry++;	/* move the pointer also */
	}
	
	/* if we didn't find it yet, start at the top of the table */
	i = 0;
	entry = &cache->table[0];
	while( i != bno ) {		/* go until we hit the bucket we started with */
		if( (entry->flags & CACHE_VALID) == CACHE_VALID ) {
			if( memcmp(&entry->ipx_header, header, sizeof(IPXHEADER)) == 0 ) {
				/* found it! */
				entry->flags          |= CACHE_ACCESSED;	/* we used this one */
				entry->access_remove   = CACHE_REMOVE_ACCESS_TICKS;	/* reset counter */
				entry->used_count++;
				return(i);
			}
		} else {
			/* found an entry which was empty */
			return(-1);
		}
		i++;		/* move to the next entry */
		entry++;	/* move along */
	}
	
	/* the cache was full, but we did not find it */
	return(-1);
}



/*
**	HDRCACHE_INSERT - insert item, returning it's number
*/
int hdrcache_insert(HDRCACHE *cache, IPXHEADER *header)
{
	int		usedmask, i, bno;
	HDRCACHE_ENTRY	*entry, newentry;
	
	/* build the new entry */
	memcpy(&newentry.ipx_header, header, sizeof(IPXHEADER));	/* copy ipx header */
	newentry.flags 		= CACHE_NEW_FLAGS;	/* default new flags */
	newentry.access_remove	= CACHE_REMOVE_ACCESS_TICKS;	/* how long till bit it removed */
	newentry.used_count	= 0;			/* has not been used */
	newentry.clear_req	= 0;
				
	/* hash the header to find a starting place */
	bno = hdrcache_hash(header) % cache->size;
	
	/* set the usedmask */
	if( cache->used == cache->size ) {	/* totally full */
		/* entry needs to be non-valid AND non-accessed */
		usedmask = (CACHE_VALID | CACHE_ACCESSED);
	} else {				/* still some empty entries */
		/* entry just needs to be non-valid */
		usedmask = (CACHE_VALID);
		cache->used++;		/* we used one more slot */
	}
	
	/* start at this bucket */
	i = bno;
	entry = &cache->table[i];
	while( i < cache->size ) {
		/* check for free entry */
		if( (entry->flags & usedmask) != usedmask ) {
			/* this item is available for use, so use it */
			memcpy(entry, &newentry, sizeof(HDRCACHE_ENTRY));
			/* return this index */
			return(i); 
		}
		
		/* increment the counter and pointer */
		i++;
		entry++;
	}
	
	/* we went to the end of the table and did not find any free slots */
	/* so start at the top and work back to "bno"                      */
	
	i = 0;
	entry = &cache->table[0];
	while( i != bno ) {	/* go until we get to where we started */
		/* check for an available slot */
		if( (entry->flags & usedmask) != usedmask ) {
			/* this item is available for use, so use it */
			memcpy(entry, &newentry, sizeof(HDRCACHE_ENTRY));
			/* return the index */
			return(i);
		}
		
		/* increment counter and pointer */
		i++;
		entry++;
	}
	
	/* everything was full and accessed, so stick it in in bno */
	entry = &cache->table[bno];
	memcpy(entry, &newentry, sizeof(HDRCACHE_ENTRY));

	/* return where we put it */
	return(bno);
}


/*
**	HDRCACHE_FLUSH - flushes the cache
*/
void hdrcache_flush(HDRCACHE *cache)
{
	/* just zero out all the memory, a memset will do fine  */
	/* not sure if this is better than looping and turning  */
	/* all the valid bits off...				*/
	memset(cache->table, 0, sizeof(HDRCACHE_ENTRY) * cache->size);
	
	/* set used number to zero */
	cache->used = 0;	

	/* done */
}


/*
**	HDRCACHE_MAINT - called peridoically to perform maintenance
*/
void hdrcache_maint(HDRCACHE *cache)
{
	int		i;
	HDRCACHE_ENTRY	*entry;
	
	/* make sure it is legal */
	if( cache == NULL )
		return;
	
	/* set the pointer to the 1st entry */
	entry = &cache->table[0];
	
	/* for each slot remove the ACCESSED bit */
	for( i = 0; i < cache->size; i++ ) {
		entry->flags &= ~CACHE_ACCESSED;	/* this entry is no longer accessed */
		
		/* move to next entry */
		entry++;
	}
	
	/* done */

}




/*
**	CREATE_HDRCACHE
*/
int create_hdrcache(HDRCACHE *cache, int size)
{
	/* clean the struct */
	memset(cache, 0, sizeof(HDRCACHE));
	
	/* alloc the memory for the table */
	cache->table = (HDRCACHE_ENTRY *)malloc(sizeof(HDRCACHE_ENTRY) * size);
	if( cache->table != NULL ) {
		/* clean memory */
		memset(cache->table, 0, sizeof(HDRCACHE_ENTRY) * size);
		/* set vars */
		cache->size = size;
		/* used was set to zero by memset */
		return(1);
	}
	
	/* failed */
	errormsg(0, "create_hdrcache", "failed to alloc memory for cache (%d entries)", size);
	return(0);
}



/*
**	FREE_HDRCACHE
*/
void free_hdrcache(HDRCACHE *cache)
{
	/* free the mem allocated */
	if( cache->table != NULL )
		free(cache->table);
		
	/* done */
}


/*
**	HDRCACHE_DUMP - dumps the cache contents
*/
void hdrcache_dump(HDRCACHE *cache)
{
	int		i;
	HDRCACHE_ENTRY	*entry;
	char		flag_buf[8];
	
	/* set ptr */
	entry = cache->table;
	
	/* setup flag buf */
	memset(flag_buf, 0, 8);
	
	/* now for each entry */
	notice("  #  Fl  A Used      Source               Destination\n");
	/*      [ 1] --- 0 12345678  001122334455:0011 -> 001122334455:0011 */
	
	for( i = 0; i < cache->size; i++ ) {		
		/* show slot */
		notice("[%2d]", i);
		/* show flags */
		memset(flag_buf, '-', 3);
		if( entry->flags & CACHE_VALID ) 
			flag_buf[0] = 'V';
		if( entry->flags & CACHE_ACCESSED )
			flag_buf[1] = 'A';
		notice(" %s", flag_buf);			
		/* show counters */
		notice(" .");
		/* show used count */
		notice(" %8d  ", entry->used_count);
		/* show addresses if valid */
		if( entry->flags & CACHE_VALID ) {
			print_ipx_addr(&entry->ipx_header.ipx_src, (PRINT_IPX_HOST | PRINT_IPX_PORT));
			notice(" -> ");
			print_ipx_addr(&entry->ipx_header.ipx_dest, (PRINT_IPX_HOST | PRINT_IPX_PORT));
		} 
		notice("\n");

#if 0		
		/* now show header */
		if( entry->flags & CACHE_VALID )
			hexdump((BYTE *)&entry->ipx_header, sizeof(IPXHEADER));
#endif
		
		/* next entry */
		entry++;
	}

	/* done */

}
