
/*
** ipxfury.c - IPX packet mover of _FURY_
**
** Keith Zaback (krz@wcug.wwu.edu) Apr 98
*/


/* includes */
#include "ipxfury.h"
#include "./version.h"
#include <stdio.h>
#include <stdarg.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/file.h>
#include <sys/param.h>


/* protos */
int			main(int argc, char *argv[]);
void			sig_handler(int signum);
void			ipx_fury(void);
int			daemonize(void);

/* globals */
GLOBALS		ifg;			/* Ipx Fury Globals */
unsigned char	broadcast_addr[] = { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF };
int		my_tty;			/* our tty */

/*
**	MAIN()
*/
int main(int argc, char *argv[])
{
	HWADDR_ENTRY	bcast;
	int		i;

	/* clean globals */
	memset(&ifg, 0, sizeof(GLOBALS));
	
	/* set the version */
	ifg.ver_major	= FURY_VER_MAJOR;
	ifg.ver_minor	= FURY_VER_MINOR;
	ifg.ver_build	= FURY_BUILD;
	
	/* do command line args */
	for( i = 1; i < argc; i++ ) {
		if( argv[i][0] == '-' ) {
			/* what option? */
			switch( argv[i][1] ) {
				case 'b':		/* BACKGROUND */
					ifg.config.background = 1;
					break;
				default:
					printf("Unknow option '%c'\n", argv[i][1]);
					return(1);
			}
		}
	}
	
	/* remove pid file */
	unlink("ipxfury.pid");	

	/* get the console going! */
	if( ifg.config.background == 0x00 ) {
		if( ! init_console(&my_tty) ) {
			printf("Could not initialize the console!\n");	
			return(1);
		}
	} else {
		/* little banner */
		printf("..::#  IPX FURY! (Ver: %d.%d.%d) Now with UDP additive! +.* Sparkly! *.+  #::..\n", ifg.ver_major, ifg.ver_minor, ifg.ver_build);
		if( ! daemonize() ) {
			return(1);
		}
	}

	/* seed the random number gen */
	srandom(time(NULL));

	
	/* set the default */
	ifg.stats.recent_stats_time 	= DEVICE_STATS_TIME;
	ifg.verbose_level		= FURY_DEFAULT_VERBOSE_LEVEL;
	
	/* set init logfile name */
	strncpy(ifg.config.logfilename, FURY_DEFAULT_LOGFILENAME, 127);
	ifg.config.logfilename[127] = 0x00;

	/* open the log file */
	ifg.logfile = fopen(ifg.config.logfilename, "w");
	if( ifg.logfile == NULL ) {
		errormsg(0, "main", "could not open logfile '%s'", ifg.config.logfilename);
		cleanup(1);
	}

	/* little message to say hi */
	notice("..::#  IPX FURY! (Ver: %d.%d.%d) Now with UDP additive! +.* Sparkly! *.+  #::..\n", ifg.ver_major, ifg.ver_minor, ifg.ver_build);

#ifdef CURSES_CONSOLE
	/* set the status bar */
	draw_statusbar("");
#endif
	
	/* install signal handlers */	
	signal(SIGINT, sig_handler);
	signal(SIGALRM, sig_handler);
	signal(SIGUSR1, sig_handler);
	signal(SIGTERM, sig_handler);

	/* ready the port table now since the configfile will mess with it */
	ifg.watched_ports = create_port_table();
	if( ifg.watched_ports == NULL ) {
		errormsg(0, "main", "could not create port table!");
		cleanup(1);
	}
	
	/* we will watch all ports to start with */
	set_all_ports(ifg.watched_ports);
		
	/* read config file */
	read_config(FURY_DEFAULT_CFG_FILENAME, &ifg.config);
	
	/* create event list */
	ifg.events = create_eventlist();
	if( ifg.events == NULL ) {
		errormsg(0, "main", "failed to create event list");
		cleanup(1);
	}
	
	/* setup interfaces */
	interface_setup();
	
	/* open all the interfaces */
	for( i = 0; i < ifg.total_interfaces; i++ ) {
		if( ! ifg.interfaces[i].open(&ifg.interfaces[i]) )
			cleanup(1);	
	}
		
	/* prepare the hash table */
	if( ! create_hash_table(&ifg.hwaddr_table,
			        FURY_HWADDR_TABLE_SIZE,
			        sizeof(HWADDR_ENTRY),
			        (int (*)(void *))ether_hash, 
			        (int (*)(void *, void *))ether_compare) ) {
		cleanup(1);
	}
	
	/* add the broadcast address to the hardware address list */
	memcpy(bcast.hwaddr, broadcast_addr, 6);
	bcast.interface = NULL;		/* no interface for this one */
	if( hash_insert(&ifg.hwaddr_table, &bcast) == NULL ) {
		errormsg(0, "main", "could not add broadcast address to hwaddr table!");
		cleanup(1);
	}
	
	/* log the starting */
	log(".xX IPXFury Starting! Xx.");

	/* go right into fury mode if we are background */
	if( ifg.config.background ) {
		ipx_fury();
		log("IPXFury Exiting...");
	} else {
		/* do the console thing */
		ifg.status = FSTATUS_PROMPT;

#ifndef CURSES_CONSOLE
		/* start them with a prompt */
		printf("IPXFury> ");
#endif
		/* main loop */
		do { 
			terminal_read(my_tty);
		
		} while( !ifg.done );
	}
		
	/* cleanup */
	cleanup(0);
	
	/* make the compiler happy */
	return(0);
}

/*
**	DAEMONIZE
*/
int daemonize()
{
	pid_t		childpid;
	int		fd;
	FILE		*pid_file;
    
	/* ignore some signals */
  	signal(SIGINT, SIG_IGN);	/* ^C */
  	signal(SIGTERM, SIG_IGN);	/* Terminate signal */
    
  	/* background ourself */
  	childpid = fork();
  	if( childpid == -1 ) {		/* failure */
    		fprintf(stderr, "daemon: Failed to fork!\n");
    		return(0);			/* failed */
  	} else if( childpid > 0 ) { 
    		exit(0);			/* the parent */
  	}
    
 	/* the child is running now */
  
#if defined FURY_LINUX
  	if( setpgrp() == -1 ) {
    		fprintf(stderr, "daemon: Failed setpgrp()\n");
	    	return(0);
  	}
#elif define FURY_BSD
  	if( setpgrp(0, getpid()) == -1 ) {
    		fprintf(stderr, "daemon: Failed setpgrp()\n");
    		return(0);
  	}
#endif

	/* detach from terminal */
  	if( (fd = open("/dev/tty", O_RDWR)) >= 0 ) {
    		ioctl(fd, TIOCNOTTY, NULL);		/* bye bye */
    		close(fd);
  	}
  
  	/* close all file descriptors */
  	for( fd = 0; fd < NOFILE; fd++ ) 
    		close(fd);
    
  	errno = 0;		/* clear it */
  	
  	/* make a pid file */
  	pid_file = fopen("ipxfury.pid", "w");
  	if( pid_file != NULL ) {
  		fprintf(pid_file, "%d\n", getpid());
  		fclose(pid_file);
  	}
  
  	return(1);		/* did ok */
}






/*
**	IPX_FURY() - the main loop of this program
*/
void ipx_fury()
{
	fd_set			readfds, writefds;
	struct timeval		select_tv;
	int			done, nfds;
	int			i, oldstatus;
	INTERFACE		*interfaces;

	
#ifdef CURSES_CONSOLE
	draw_statusbar(" (( FURIOUS! [^D to quit] ))");
#endif
	
	/* set the status */
	oldstatus = ifg.status;
	ifg.status = FSTATUS_FURIOUS;				
	
	/* init variables */
	interfaces 	= ifg.interfaces;
	
	/* flush all interfaces & prepare them */
	for( i = 0; i < ifg.total_interfaces; i++ ) {
		interfaces[i].flush(&interfaces[i]);
		interfaces[i].prepare_device(&interfaces[i]);
	}
		
	/* start the cache maintence event */
	add_event(ifg.events,
	          ETYPE_CACHE_MAINT,
	          CACHE_MAINTENCE_TIME,
	          CACHE_MAINTENCE_TIME,
	          NULL, NULL);
	          
	/* add in the stats events -- when not in background mode */
	if( ifg.config.background == 0x00 ) {
		add_event(ifg.events, 
		          ETYPE_PERIODIC_STATS,
		          ifg.stats.recent_stats_time,
		          ifg.stats.recent_stats_time,
		          NULL, NULL);			
	}
	
	/* if they have a UDP interface, add in the udp_maint */
	if( NULL != interface_find_name("udp") ) {
		add_event(ifg.events,
			  ETYPE_UDP_MAINT,
			  UDP_MAINTENCE_TIME,
			  UDP_MAINTENCE_TIME,
			  NULL, NULL);
	}
	
	/* the main loop */
	done = 0;
	while( (!done) && (!ifg.done) ) {
		/* clean out fdsets */
		FD_ZERO(&readfds);
		FD_ZERO(&writefds);
		
		/* set the files we want to watch */
		for( i = 0; i < ifg.total_interfaces; i++ ) {
			FD_SET(interfaces[i].fd, &readfds);
			if( buffer_dirty(&interfaces[i].sendbuf) ) {
				FD_SET(interfaces[i].fd, &writefds);
			}
		}
		
		/* put the terminal in the read group -- if not background */
		if( ifg.config.background == 0 ) {
			FD_SET(my_tty, &readfds);
		}
		
		/* set select timeout */
		select_tv.tv_sec = FURY_WAIT_SECONDS;
		select_tv.tv_usec = FURY_WAIT_USECONDS;
		
		/* now select() */
		/* 11/6 - ignoring writefds .. */
		nfds = select(FD_SETSIZE, &readfds, NULL, NULL, &select_tv);

		/* set the now timer */
		gettimeofday(&ifg.now, NULL);
		
		/* do the proper thing if there is something to read or write */
		if( nfds >= 0 ) {
						
			/* -- TERMINAL INPUT -- */
			if( ifg.config.background == 0 ) {
				if( FD_ISSET(my_tty, &readfds) ) {
					done = terminal_read(my_tty);
					if( done == 1 ) {
						/* try a good exit */
						for( i = 0; i < ifg.total_interfaces; i++ ) {
							interfaces[i].event_handler(&interfaces[i], FURY_EVENT_CLOSE);
						}
						done = 0;
					} else if( done == 2 ) {
						break;		/* really quit */
					}
				} /* end of terminal input */
			}

			/* -- INTERFACE INPUT -- */
			/* this also processes and writes.. and		*/
			/* checks for closed interfaces			*/
			done = 1;
			for( i = 0; i < ifg.total_interfaces; i++ ) {		
				if( FD_ISSET(interfaces[i].fd, &readfds) ) {				
					interfaces[i].read(&interfaces[i]);
				}
				/* always process */
				interfaces[i].process(&interfaces[i]);
				
				/* now write */
				interfaces[i].write(&interfaces[i]);
				
				/* is it closed? */
				if( interfaces[i].state > FURY_STATE_CLOSED )
					done = 0;
			}
			
		} /* if select() >= 0 */
		
		/* do all the events */
		do_events(ifg.events, &ifg.now);

		/* -- possible carrier check for serial device -- */
	
	} /* while */

	/* clean out the events */
	event_clean(ifg.events);

	/* force the interfaces into an INITIAL state */
	for( i = 0; i < ifg.total_interfaces; i++ ) {
		interfaces[i].state = FURY_STATE_INITIAL;
	}

	/* change the status back */
	ifg.status = oldstatus;
	
#ifdef CURSES_CONSOLE
	/* blank the status bar */
 	draw_statusbar("");
#endif

	/* done */
	return;
}



/*
**	CLEANUP()
*/
void cleanup(int exit_code)
{
	int		i;
	
	/* free all interfaces */
	for( i = 0; i < ifg.total_interfaces; i++ ) {
		ifg.interfaces[i].close(&ifg.interfaces[i]);
	}

	/* free the hash table */		
	free_hash_table(&ifg.hwaddr_table);

	/* free the event list */
	free_eventlist(ifg.events);
	
	/* free the port table */
	free_port_table(&ifg.watched_ports);

	/* close the logfile */
	if( ifg.logfile != NULL )
		fclose(ifg.logfile);
		
	/* exit the console */
	if( ifg.config.background == 0x00 ) {
		exit_console(my_tty);
	}
	
	/* remove pid file */
	unlink("ipxfury.pid");	
	
	/* we are done */
	exit(exit_code);

}


/*
**	SIG_HANDLER()
*/
void sig_handler(int signum)
{
	int		i;	
	
	/* process the signal */
	switch( signum ) {
		case SIGTERM:
			log("SIGTERM: Closing interfaces.");
			for( i = 0; i < ifg.total_interfaces; i++ ) {
				ifg.interfaces[i].event_handler(&ifg.interfaces[i], FURY_EVENT_CLOSE);
			}
			break;
			
		case SIGINT:		/* ^C */
			log("SIGINT: Exiting!!");
			cleanup(1);
			break;
			
		case SIGUSR1:		/* stats */
			show_all_stats();
			break;
			
		case SIGALRM:
			/* set alarm flag */
			ifg.alarm = 1;
			break;
	}
	
	/* restore the handler */
	signal(signum, sig_handler);
	
}


